@@ -16,6 +16,7 @@ def test_with_query():
 def test_update_query():
     url = URL('http://example.com/')
     assert str(url.update_query({'a': '1'})) == 'http://example.com/?a=1'
+    assert str(URL('test').update_query(a=1)) == 'test?a=1'
 
     url = URL('http://example.com/?foo=bar')
     expected_url = URL('http://example.com/?foo=bar&baz=foo')
@@ -68,7 +69,9 @@ def test_with_query_only_single_arg_is_supported():
 
 def test_with_query_empty_dict():
     url = URL('http://example.com/?a=b')
-    assert str(url.with_query({})) == 'http://example.com/'
+    new_url = url.with_query({})
+    assert new_url.query_string == ''
+    assert str(new_url) == 'http://example.com/'
 
 
 def test_with_query_empty_str():
