@@ -1,3 +1,5 @@
+import warnings
+from collections import OrderedDict
 from collections.abc import Mapping, Sequence
 from functools import partial
 from ipaddress import ip_address
@@ -9,13 +11,12 @@ from multidict import MultiDict, MultiDictProxy
 
 from .quoting import quote, unquote
 
-__version__ = '0.8.1'
+__version__ = '0.16.0'
 
-__all__ = ['URL']
+__all__ = ('URL',)
 
 
 # is_leaf()
-# path normalization
 
 
 DEFAULT_PORTS = {
@@ -25,6 +26,7 @@ DEFAULT_PORTS = {
     'wss': 443,
 }
 
+PROTECT_CHARS = '=+&;'
 
 sentinel = object()
 
@@ -45,7 +47,7 @@ class cached_property:
         self.wrapped = wrapped
         try:
             self.__doc__ = wrapped.__doc__
-        except:  # pragma: no cover
+        except AttributeError:  # pragma: no cover
             self.__doc__ = ""
         self.name = wrapped.__name__
 
@@ -135,14 +137,12 @@ class URL:
     # absolute-URI  = scheme ":" hier-part [ "?" query ]
     __slots__ = ('_cache', '_val')
 
-    def __new__(cls, val='', *, encoded=False):
-        if isinstance(val, URL):
-            return val
-        else:
-            return super(URL, cls).__new__(cls)
-
-    def __init__(self, val='', *, encoded=False):
+    def __init__(self, val='', *, encoded=False, strict=None):
+        if strict is not None:  # pragma: no cover
+            warnings.warn("strict parameter is ignored")
         if isinstance(val, URL):
+            self._val = val._val
+            self._cache = val._cache
             return
         if isinstance(val, str):
             val = urlsplit(val)
@@ -167,29 +167,70 @@ class URL:
                 else:
                     try:
                         ip = ip_address(netloc)
-                    except:
+                    except ValueError:
                         pass
                     else:
                         if ip.version == 6:
-                            netloc = '['+netloc+']'
+                            netloc = '[' + netloc + ']'
                 if val.port:
                     netloc += ':{}'.format(val.port)
                 if val.username:
                     user = _quote(val.username)
-                    if val.password:
-                        user += ':' + _quote(val.password)
+                else:
+                    user = ''
+                if val.password:
+                    user += ':' + _quote(val.password)
+                if user:
                     netloc = user + '@' + netloc
 
-            val = SplitResult(val[0],  # scheme
-                              netloc,
-                              _quote(val[2], safe='@:', protected='/'),
-                              query=_quote(val[3], safe='=+&?/:@',
-                                           protected='=+&', qs=True),
-                              fragment=_quote(val[4], safe='?/:@'))
+            path = _quote(val[2], safe='+@:', protected='/+')
+            if netloc:
+                path = _normalize_path(path)
+
+            query = _quote(val[3], safe='=+&?/:@',
+                           protected=PROTECT_CHARS, qs=True)
+            fragment = _quote(val[4], safe='?/:@')
+            val = SplitResult(val[0], netloc, path, query, fragment)
 
         self._val = val
         self._cache = {}
 
+    @classmethod
+    def build(cls, *, scheme='', user='', password='', host='', port=None,
+              path='', query=None, query_string='', fragment=''):
+        """Creates and returns a new URL"""
+
+        if host and not scheme:
+            raise ValueError(
+                'Can\'t build URL with "host" but without "scheme".')
+        if not host and scheme:
+            raise ValueError(
+                'Can\'t build URL with "scheme" but without "host".')
+        if query and query_string:
+            raise ValueError(
+                "Only one of \"query\" or \"query_string\" should be passed")
+
+        netloc = cls._make_netloc(user, password, host, port)
+        path = _quote(path, safe='@:', protected='/')
+        if netloc:
+            path = _normalize_path(path)
+
+        url = cls(
+            SplitResult(
+                scheme,
+                netloc,
+                path,
+                _quote(query_string),
+                fragment
+            ),
+            encoded=True
+        )
+
+        if query:
+            return url.with_query(query)
+        else:
+            return url
+
     def __str__(self):
         val = self._val
         if not val.path and self.is_absolute() and (val.query or val.fragment):
@@ -202,12 +243,24 @@ class URL:
     def __eq__(self, other):
         if not isinstance(other, URL):
             return NotImplemented
-        return self._val == other._val
+
+        val1 = self._val
+        if not val1.path and self.is_absolute():
+            val1 = val1._replace(path='/')
+
+        val2 = other._val
+        if not val2.path and other.is_absolute():
+            val2 = val2._replace(path='/')
+
+        return val1 == val2
 
     def __hash__(self):
         ret = self._cache.get('hash')
         if ret is None:
-            ret = self._cache['hash'] = hash(self._val)
+            val = self._val
+            if not val.path and self.is_absolute():
+                val = val._replace(path='/')
+            ret = self._cache['hash'] = hash(val)
         return ret
 
     def __le__(self, other):
@@ -244,9 +297,22 @@ class URL:
             parts = path.rstrip('/').split('/')
             parts.append(name)
             new_path = '/'.join(parts)
+        if self.is_absolute():
+            new_path = _normalize_path(new_path)
         return URL(self._val._replace(path=new_path, query='', fragment=''),
                    encoded=True)
 
+    def __getstate__(self):
+        return self._val,
+
+    def __setstate__(self, state):
+        if state[0] is None and isinstance(state[1], dict):
+            # default style pickle
+            self._val = state[1]['_val']
+        else:
+            self._val, *unused = state
+        self._cache = {}
+
     def is_absolute(self):
         """A check for absolute URLs.
 
@@ -315,7 +381,10 @@ class URL:
 
         """
         # not .username
-        return self._val.username
+        ret = self._val.username
+        if not ret:
+            return None
+        return ret
 
     @cached_property
     def user(self):
@@ -396,7 +465,7 @@ class URL:
         / for absolute URLs without path part.
 
         """
-        return _unquote(self.raw_path)
+        return _unquote(self.raw_path, unsafe='+')
 
     @cached_property
     def query(self):
@@ -406,7 +475,8 @@ class URL:
         Empty value if URL has no query part.
 
         """
-        ret = MultiDict(parse_qsl(self.query_string, keep_blank_values=True))
+        ret = MultiDict(parse_qsl(self.raw_query_string,
+                                  keep_blank_values=True))
         return MultiDictProxy(ret)
 
     @property
@@ -427,6 +497,20 @@ class URL:
         """
         return _unquote(self.raw_query_string, qs=True)
 
+    @cached_property
+    def path_qs(self):
+        """Decoded path of URL with query."""
+        if not self.query_string:
+            return self.path
+        return '{}?{}'.format(self.path, self.query_string)
+
+    @cached_property
+    def raw_path_qs(self):
+        """Encoded path of URL with query."""
+        if not self.raw_query_string:
+            return self.raw_path
+        return '{}?{}'.format(self.raw_path, self.raw_query_string)
+
     @property
     def raw_fragment(self):
         """Encoded fragment part of URL.
@@ -516,7 +600,7 @@ class URL:
             ret = ret + ':' + str(port)
         if password:
             if not user:
-                raise ValueError("Non-empty password requires non-empty user")
+                user = ''
             user = user + ':' + password
         if user:
             ret = user + '@' + ret
@@ -530,8 +614,7 @@ class URL:
         if not self.is_absolute():
             raise ValueError("scheme replacement is not allowed "
                              "for relative URLs")
-        return URL(self._val._replace(scheme=scheme.lower()),
-                   encoded=True)
+        return URL(self._val._replace(scheme=scheme.lower()), encoded=True)
 
     def with_user(self, user):
         """Return a new URL with user replaced.
@@ -553,13 +636,11 @@ class URL:
         if not self.is_absolute():
             raise ValueError("user replacement is not allowed "
                              "for relative URLs")
-        return URL(
-            self._val._replace(
-                netloc=self._make_netloc(user,
-                                         password,
-                                         val.hostname,
-                                         val.port)),
-            encoded=True)
+        return URL(self._val._replace(netloc=self._make_netloc(user,
+                                                               password,
+                                                               val.hostname,
+                                                               val.port)),
+                   encoded=True)
 
     def with_password(self, password):
         """Return a new URL with password replaced.
@@ -607,17 +688,18 @@ class URL:
             raise ValueError("host removing is not allowed")
         try:
             ip = ip_address(host)
-        except:
+        except ValueError:
             host = host.encode('idna').decode('ascii')
         else:
             if ip.version == 6:
-                host = '['+host+']'
+                host = '[' + host + ']'
         val = self._val
-        return URL(self._val._replace(netloc=self._make_netloc(val.username,
-                                                               val.password,
-                                                               host,
-                                                               val.port)),
-                   encoded=True)
+        return URL(
+            self._val._replace(netloc=self._make_netloc(val.username,
+                                                        val.password,
+                                                        host,
+                                                        val.port)),
+            encoded=True)
 
     def with_port(self, port):
         """Return a new URL with port replaced.
@@ -633,25 +715,25 @@ class URL:
             raise ValueError("port replacement is not allowed "
                              "for relative URLs")
         val = self._val
-        return URL(self._val._replace(netloc=self._make_netloc(val.username,
-                                                               val.password,
-                                                               val.hostname,
-                                                               port)),
+        return URL(
+            self._val._replace(netloc=self._make_netloc(val.username,
+                                                        val.password,
+                                                        val.hostname,
+                                                        port)),
+            encoded=True)
+
+    def with_path(self, path, *, encoded=False):
+        """Return a new URL with path replaced."""
+        if not encoded:
+            path = _quote(path, safe='@:', protected='/')
+            if self.is_absolute():
+                path = _normalize_path(path)
+        if len(path) > 0 and path[0] != '/':
+            path = '/' + path
+        return URL(self._val._replace(path=path, query='', fragment=''),
                    encoded=True)
 
-    def with_query(self, *args, **kwargs):
-        """Return a new URL with query part replaced.
-
-        Accepts any Mapping (e.g. dict, multidict.MultiDict instances)
-        or str, autoencode the argument if needed.
-
-        It also can take an arbitrary number of keyword arguments.
-
-        Clear query if None is passed.
-
-        """
-        # N.B. doesn't cleanup query/fragment
-
+    def _get_str_query(self, *args, **kwargs):
         if kwargs:
             if len(args) > 0:
                 raise ValueError("Either kwargs or single query parameter "
@@ -666,7 +748,7 @@ class URL:
         if query is None:
             query = ''
         elif isinstance(query, Mapping):
-            quoter = partial(_quote, safe='/?:@', qs=True)
+            quoter = partial(_quote, qs=True)
             lst = []
             for k, v in query.items():
                 if isinstance(v, str):
@@ -676,25 +758,59 @@ class URL:
                 else:
                     raise TypeError("Invalid variable type: mapping value "
                                     "should be str or int, got {!r}".format(v))
-                lst.append(quoter(k)+'='+quoter(v))
+                lst.append(
+                    quoter(k, safe='/?:@') + '=' + quoter(v, safe='/?:@'))
             query = '&'.join(lst)
         elif isinstance(query, str):
-            query = _quote(query, safe='/?:@', protected='=&+', qs=True)
+            query = _quote(query, safe='/?:@',
+                           protected=PROTECT_CHARS,
+                           qs=True)
         elif isinstance(query, (bytes, bytearray, memoryview)):
             raise TypeError("Invalid query type: bytes, bytearray and "
                             "memoryview are forbidden")
         elif isinstance(query, Sequence):
-            quoter = partial(_quote, safe='/?:@', qs=True)
-            query = '&'.join(quoter(k)+'='+quoter(v)
+            quoter = partial(_quote, qs=True, safe='/?:@')
+            query = '&'.join(quoter(k) + '=' + quoter(v)
                              for k, v in query)
         else:
             raise TypeError("Invalid query type: only str, mapping or "
                             "sequence of (str, str) pairs is allowed")
-        path = self._val.path
-        if path == '':
-            path = '/'
-        return URL(self._val._replace(path=path, query=query),
-                   encoded=True)
+
+        return query
+
+    def with_query(self, *args, **kwargs):
+        """Return a new URL with query part replaced.
+
+        Accepts any Mapping (e.g. dict, multidict.MultiDict instances)
+        or str, autoencode the argument if needed.
+
+        It also can take an arbitrary number of keyword arguments.
+
+        Clear query if None is passed.
+
+        """
+        # N.B. doesn't cleanup query/fragment
+
+        new_query = self._get_str_query(*args, **kwargs)
+        return URL(
+            self._val._replace(path=self._val.path, query=new_query),
+            encoded=True)
+
+    def update_query(self, *args, **kwargs):
+        """Return a new URL with query part updated."""
+        new_query = OrderedDict(
+            map(
+                lambda x: x.split('=', 1),
+                _quote(self._get_str_query(*args, **kwargs),
+                       safe='/?:@', protected=PROTECT_CHARS,
+                       qs=True).lstrip("?").split("&")
+            )
+        )
+        query = OrderedDict(self.query)
+        query.update(new_query)
+        query = self._get_str_query(query)
+        return URL(
+            self._val._replace(path=self._val.path, query=query), encoded=True)
 
     def with_fragment(self, fragment):
         """Return a new URL with fragment replaced.
@@ -709,8 +825,10 @@ class URL:
             fragment = ''
         elif not isinstance(fragment, str):
             raise TypeError("Invalid fragment type")
-        return URL(self._val._replace(fragment=_quote(fragment, safe='?/:@')),
-                   encoded=True)
+        return URL(
+            self._val._replace(
+                fragment=_quote(fragment, safe='?/:@')),
+            encoded=True)
 
     def with_name(self, name):
         """Return a new URL with name (last part of path) replaced.
@@ -726,6 +844,8 @@ class URL:
         if '/' in name:
             raise ValueError("Slash in name is not allowed")
         name = _quote(name, safe='@:', protected='/')
+        if name in ('.', '..'):
+            raise ValueError(". and .. values are forbidden")
         parts = list(self.raw_parts)
         if self.is_absolute():
             if len(parts) == 1:
@@ -737,8 +857,9 @@ class URL:
             parts[-1] = name
             if parts[0] == '/':
                 parts[0] = ''  # replace leading '/'
-        return URL(self._val._replace(path='/'.join(parts),
-                                      query='', fragment=''), encoded=True)
+        return URL(
+            self._val._replace(path='/'.join(parts),
+                               query='', fragment=''), encoded=True)
 
     def join(self, url):
         """Join URLs
@@ -768,3 +889,31 @@ class URL:
                                       self.path,
                                       self.query_string,
                                       self.fragment))
+
+
+def _normalize_path(path):
+    # Drop '.' and '..' from path
+
+    segments = path.split('/')
+    resolved_path = []
+
+    for seg in segments:
+        if seg == '..':
+            try:
+                resolved_path.pop()
+            except IndexError:
+                # ignore any .. segments that would otherwise cause an
+                # IndexError when popped from resolved_path if
+                # resolving for rfc3986
+                pass
+        elif seg == '.':
+            continue
+        else:
+            resolved_path.append(seg)
+
+    if segments[-1] in ('.', '..'):
+        # do some post-processing here. if the last segment was a relative dir,
+        # then we need to append the trailing '/'
+        resolved_path.append('')
+
+    return '/'.join(resolved_path)
